#!/usr/bin/env python
# Copyright 2008, 2009 Hannes Hochreiner
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# These lines are only needed if you don't put the script directly into
# the installation directory
import sys
# Unix
sys.path.append('/usr/share/inkscape/extensions')
# OS X
sys.path.append('/Applications/Inkscape.app/Contents/Resources/extensions')
# Windows
sys.path.append('C:\Program Files\Inkscape\share\extensions')

import inkex, os.path
import subprocess
import tempfile
import os
import zipfile
import glob

def propStrToDict(inStr):
	dictio = {}

	for prop in inStr.split(";"):
		values = prop.split(":")

		if (len(values) == 2):
			dictio[values[0].strip()] = values[1].strip()

	return dictio

def dictToPropStr(dictio):
	str = ""

	for key in dictio.keys():
		str += " " + key + ":" + dictio[key] + ";" 

	return str[1:]

def setStyle(node, propKey, propValue):
	props = {}

	if node.attrib.has_key("style"):
		props = propStrToDict(node.get("style"))
	
	props[propKey] = propValue
	node.set("style", dictToPropStr(props))

class MyEffect(inkex.Effect):
	inkscapeCommand = None
	zipFile = None

	def __init__(self):
		inkex.Effect.__init__(self)

		# Register jessyink namespace.
		inkex.NSS[u"jessyink"] = u"https://launchpad.net/jessyink"

		# Set inkscape command.
		self.inkscapeCommand = self.findInkscapeCommand()

		if (self.inkscapeCommand == None):
			sys.stderr.write("Could not find Inkscape command.\n")
			sys.exit(1)

	def output(self):
		pass

	def effect(self):
		# Check version.
		scriptNodes = self.document.xpath("//svg:script[@jessyink:version='1.3.1']", namespaces=inkex.NSS)

		if len(scriptNodes) != 1:
			sys.stderr.write("The JessyInk script is not installed in this SVG file or has a different version than the JessyInk extensions. Please select \"install/update...\" from the \"JessyInk\" sub-menu of the \"Effects\" menu to install or update the JessyInk script.\n\n")

		# Remove any temporary files that might be left from last time.
		self.removeJessyInkFilesInTempDir()

		zipFileDesc, zpFile = tempfile.mkstemp(suffix=".zip", prefix="jessyInk__")

		output = zipfile.ZipFile(zpFile, "w", compression=zipfile.ZIP_STORED)
		slides = []
		masterSlide = None

		# Find width and height of image.
		self.imageWidth = float(self.document.xpath("/svg:svg", namespaces=inkex.NSS)[0].get("width"));
		self.imageHeight = float(self.document.xpath("/svg:svg", namespaces=inkex.NSS)[0].get("height"));

		# Make slide array and find master slide (if assigned).
		nodes = self.document.xpath("//*[@inkscape:groupmode='layer']", namespaces=inkex.NSS)

		for node in nodes:
			if node.attrib.has_key("{" + inkex.NSS["jessyink"] + "}masterSlide"):
				masterSlide = node

				# Set master slide visibility.
				setStyle(node, "display", "inherit")
			else:
				# Set visibility.
				setStyle(node, "display", "none")
				setStyle(node, "opacity", "0")
				slides.append(node)

		# Remove master slide from document.
		if masterSlide != None:
			masterSlide.getparent().remove(masterSlide)

		# Set slide number.
		slideNumber = 1

		# Iterate through slides.
		for slide in slides:
			# Prepare svg file.
			setStyle(slide, "display", "inherit")
			setStyle(slide, "opacity", "1")

			# Insert master slide.
			if masterSlide != None:
				slide.insert(0, masterSlide)

			# Substitute auto-texts.
			autoTextNodes = slide.xpath(".//*[@jessyink:autoText]", namespaces=inkex.NSS)

			for autoTextNode in autoTextNodes:
				autoTextKind = autoTextNode.get("{" + inkex.NSS["jessyink"] + "}autoText")

				if autoTextKind == "slideTitle":
					autoTextNode.text = slide.get("{" + inkex.NSS["inkscape"] + "}label")
				elif autoTextKind == "slideNumber":
					autoTextNode.text = str(slideNumber)
				elif autoTextKind == "numberOfSlides":
					autoTextNode.text = str(len(slides))

			# Collect information about effects.
			effects = {}

			for effectNode in slide.xpath(".//*[@jessyink:effectIn]", namespaces=inkex.NSS):
				dictio = propStrToDict(effectNode.get("{" + inkex.NSS["jessyink"] + "}effectIn"))
				dictio["direction"] = "in"
				dictio["node"] = effectNode
				dictio["type"] = "effect"

				if not effects.has_key(dictio["order"]):
					effects[dictio["order"]] = []

				effects[dictio["order"]].append(dictio)

			for effectNode in slide.xpath(".//*[@jessyink:effectOut]", namespaces=inkex.NSS):
				dictio = propStrToDict(effectNode.get("{" + inkex.NSS["jessyink"] + "}effectOut"))
				dictio["direction"] = "out"
				dictio["node"] = effectNode
				dictio["type"] = "effect"

				if not effects.has_key(dictio["order"]):
					effects[dictio["order"]] = []

				effects[dictio["order"]].append(dictio)

			# Collect information about views.
			matrixOld = matrixSVG.fromElements(1, 0, 0, 0, 1, 0, 0, 0, 1)
			tempViews = {}

			# Find views.
			for viewNode in slide.xpath(".//*[@jessyink:view]", namespaces=inkex.NSS):
				dictio = propStrToDict(viewNode.get("{" + inkex.NSS["jessyink"] + "}view"))
				
				if float(dictio["order"]) == 0:
					matrixOld = self.pointMatrixToTransformation(self.rectToMatrix(viewNode))
				else:
					dictio = propStrToDict(viewNode.get("{" + inkex.NSS["jessyink"] + "}view"))
					dictio["node"] = slide
					dictio["type"] = "view"
					dictio["matrixNew"] = self.pointMatrixToTransformation(self.rectToMatrix(viewNode))

					tempViews[dictio["order"]] = dictio

				# Remove node.
				viewNode.getparent().remove(viewNode)

			slide.set("transform", matrixOld.toAttribute())

			# Consolidate views and add them to effects.
			orderViews = sorted(tempViews.keys())

			for viewNumber in orderViews:
				tempViews[viewNumber]["matrixOld"] = matrixOld
				matrixOld = tempViews[viewNumber]["matrixNew"]

				if not effects.has_key(tempViews[viewNumber]["order"]):
					effects[tempViews[viewNumber]["order"]] = []

				effects[tempViews[viewNumber]["order"]].append(tempViews[viewNumber])

			order = sorted(effects.keys())
			order.reverse()

			# Set slide to initial state.
			for effectNumber in order:
				for effect in effects[effectNumber]:
					if effect["type"] == "effect":
						if effect["direction"] == "out":
							setStyle(effect["node"], "display", "inherit")
							setStyle(effect["node"], "opacity", "1")
						elif effect["direction"] == "in":
							setStyle(effect["node"], "display", "none")
							setStyle(effect["node"], "opacity", "0")
					elif effect["type"] == "view":
						effect["node"].set("transform", effect["matrixOld"].toAttribute())

			# Add new page.
			effectNmb = 0
			self.takeSnapshot(output, slideNumber, len(slides), effectNmb, len(order) - 1)

			# Step through effects.
			order.reverse()

			for effectNumber in order:
				effectNmb += 1

				for effect in effects[effectNumber]:
					if effect["type"] == "effect":
						if effect["direction"] == "in":
							setStyle(effect["node"], "display", "inherit")
							setStyle(effect["node"], "opacity", "1")
						elif effect["direction"] == "out":
							setStyle(effect["node"], "display", "none")
							setStyle(effect["node"], "opacity", "0")
					elif effect["type"] == "view":
						effect["node"].set("transform", effect["matrixNew"].toAttribute())

				self.takeSnapshot(output, slideNumber, len(slides), effectNmb, len(order) - 1)

			# Post-process svg.
			setStyle(slide, "display", "none")
			setStyle(slide, "opacity", "0")

			# Increment slide number.
			slideNumber += 1
				
		# Write temporary zip file to stdout.
		output.close()
		out = open(zpFile,'rb')

		# Switch stdout to binary on Windows.
		if sys.platform == "win32":
			import os, msvcrt
			msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

		# Output the file.
		sys.stdout.write(out.read())
		sys.stdout.close()
		out.close()

		# Delete temporary files.
		self.removeJessyInkFilesInTempDir()

	# Function to export the current state of the file using Inkscape.
	def takeSnapshot(self, output, slideNumber, maxSlideNumber, effectNumber, maxEffectNumber):
		# Write the svg file.
		svgFileDesc, svgFile = tempfile.mkstemp(suffix=".svg", prefix="jessyInk__")
		self.document.write(os.fdopen(svgFileDesc, "wb"))
		
		# Prepare pdf file.
		pdfFileDesc, pdfFile = tempfile.mkstemp(suffix=".pdf", prefix="jessyInk__")
				
		proc = subprocess.Popen([self.inkscapeCommand + " --file=" + svgFile  + " --without-gui --export-pdf=" + pdfFile], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		stdout_value, stderr_value = proc.communicate()
		
		output.write(pdfFile, "slide_" + self.padNumber(slideNumber, maxSlideNumber) + "_effect_" + self.padNumber(effectNumber, maxEffectNumber) + ".pdf")

	# Function to remove any temporary files created during the export.
	def removeJessyInkFilesInTempDir(self):
		for infile in glob.glob(os.path.join(tempfile.gettempdir(), 'jessyInk__*')):
			try:
				os.remove(infile)
			except:
				pass
				
	# Function to prepend zero to a number in order to ensure that the numbers are all the same length.
	def padNumber(self, nmb, maxNmb):
		if (maxNmb < 0):
			maxNmb = 0

		maxNmbStr = str(maxNmb)
		nmbStr = str(nmb)
		
		while len(nmbStr) < len(maxNmbStr):
			nmbStr = "0" + nmbStr

		return nmbStr

	# Function to try and find the correct command to invoke Inkscape.
	def findInkscapeCommand(self):
		commands = []
		commands.append("inkscape")
		commands.append("C:\Program Files\Inkscape\inkscape.exe")
		commands.append("/Applications/Inkscape.app/Contents/Resources/bin/inkscape")

		for command in commands:
			proc = subprocess.Popen([command + " --without-gui --version"], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			stdout_value, stderr_value = proc.communicate()

			if proc.returncode == 0:
				return command

		return None

	# Convenience function to convert a point matrix into a transformation matrix.
	def pointMatrixToTransformation(self, mPoints):
		mPointsOld = matrixSVG.fromElements(0, self.imageWidth, self.imageWidth, 0, 0, self.imageHeight, 1, 1, 1)
	
		return mPointsOld * ~mPoints

	# Convenience function to obtain a point matrix from a rectangle.
	def rectToMatrix(self, rect):
		rectWidth = float(rect.get("width"))
		rectHeight = float(rect.get("height"))
		rectX = float(rect.get("x"))
		rectY = float(rect.get("y"))
		rectXcorr = 0
		rectYcorr = 0
	
		scaleX = self.imageWidth / rectWidth
		scaleY = self.imageHeight / rectHeight
	
		if (scaleX > scaleY):
			scaleX = scaleY
			rectXcorr -= (self.imageWidth / scaleX - rectWidth) / 2
			rectWidth = self.imageWidth / scaleX
		else:
			scaleY = scaleX
			rectYcorr -= (self.imageHeight / scaleY - rectHeight) / 2
			rectHeight = self.imageHeight / scaleY
	
		mRectTrans = matrixSVG.fromElements(1, 0, 0, 0, 1, 0, 0, 0, 1)
	
		if (rect.get("transform")):
			mRectTrans = matrixSVG.fromAttribute(rect.get("transform"))
	
		newBasePoints = matrixSVG.fromElements(rectX, rectX, rectX, rectY, rectY, rectY, 1, 1, 1)
		newVectors = matrixSVG.fromElements(rectXcorr, rectXcorr + rectWidth, rectXcorr + rectWidth, rectYcorr, rectYcorr, rectYcorr + rectHeight, 0, 0, 0)
	
		return mRectTrans * (newBasePoints + newVectors);

# Class representing an SVG transformation matrix.
class matrixSVG:
	e11 = 0.0
	e12 = 0.0
	e13 = 0.0
	e21 = 0.0
	e22 = 0.0
	e23 = 0.0
	e31 = 0.0
	e32 = 0.0
	e33 = 0.0

	# Contructor.
	@classmethod
	def fromElements(self, e11, e12, e13, e21, e22, e23, e31, e32, e33):
		out = matrixSVG()

		out.e11 = e11
		out.e12 = e12
		out.e13 = e13
		out.e21 = e21
		out.e22 = e22
		out.e23 = e23
		out.e31 = e31
		out.e32 = e32
		out.e33 = e33

		return out

	# Contructor.
	@classmethod
	def fromAttribute(self, attrString):
		out = matrixSVG()
		str = attrString[7:-1]
	
		strArray = str.split(",")
	
		out.e11 = float(strArray[0])
		out.e21 = float(strArray[1])
		out.e31 = 0
		out.e12 = float(strArray[2])
		out.e22 = float(strArray[3])
		out.e32 = 0
		out.e13 = float(strArray[4])
		out.e23 = float(strArray[5])
		out.e33 = 1
		
		return out

	# Convert 3x3 matrix into an SVG transformation matrix.
	def toAttribute(self):
		return "matrix(" + str(self.e11) + ", " + str(self.e21) + ", " + str(self.e12) + ", " + str(self.e22) + ", " + str(self.e13) + ", " + str(self.e23) + ")";

	# Inversion the matrix.
	def __invert__(self):
		out = matrixSVG()
	
		det = self.e11 * (self.e33 * self.e22 - self.e32 * self.e23) - self.e21 * (self.e33 * self.e12 - self.e32 * self.e13) + self.e31 * (self.e23 * self.e12 - self.e22 * self.e13)
	
		out.e11 =  (self.e33 * self.e22 - self.e32 * self.e23) / det
		out.e12 = -(self.e33 * self.e12 - self.e32 * self.e13) / det
		out.e13 =  (self.e23 * self.e12 - self.e22 * self.e13) / det
		out.e21 = -(self.e33 * self.e21 - self.e31 * self.e23) / det
		out.e22 =  (self.e33 * self.e11 - self.e31 * self.e13) / det
		out.e23 = -(self.e23 * self.e11 - self.e21 * self.e13) / det
		out.e31 =  (self.e32 * self.e21 - self.e31 * self.e22) / det
		out.e32 = -(self.e32 * self.e11 - self.e31 * self.e12) / det
		out.e33 =  (self.e22 * self.e11 - self.e21 * self.e12) / det
	
		return out

	# Multiplication with another matrix.
	def __mul__(self, op):
		out = matrixSVG()
	
		out.e11 = self.e11 * op.e11 + self.e12 * op.e21 + self.e13 * op.e31
		out.e12 = self.e11 * op.e12 + self.e12 * op.e22 + self.e13 * op.e32
		out.e13 = self.e11 * op.e13 + self.e12 * op.e23 + self.e13 * op.e33
		out.e21 = self.e21 * op.e11 + self.e22 * op.e21 + self.e23 * op.e31
		out.e22 = self.e21 * op.e12 + self.e22 * op.e22 + self.e23 * op.e32
		out.e23 = self.e21 * op.e13 + self.e22 * op.e23 + self.e23 * op.e33
		out.e31 = self.e31 * op.e11 + self.e32 * op.e21 + self.e33 * op.e31
		out.e32 = self.e31 * op.e12 + self.e32 * op.e22 + self.e33 * op.e32
		out.e33 = self.e31 * op.e13 + self.e32 * op.e23 + self.e33 * op.e33
	
		return out

	# Addition of another matrix.
	def __add__(self, op):
		out = matrixSVG()
	
		out.e11 = self.e11 + op.e11
		out.e12 = self.e12 + op.e12
		out.e13 = self.e13 + op.e13
		out.e21 = self.e21 + op.e21
		out.e22 = self.e22 + op.e22
		out.e23 = self.e23 + op.e23
		out.e31 = self.e31 + op.e31
		out.e32 = self.e32 + op.e32
		out.e33 = self.e33 + op.e33
	
		return out

	# Mixing of matrices (1 - f) * X + f * Y.
	def mix(self, op, contribOp):
		contribThis = 1.0 - contribOp
		out = matrixSVG()
	
		out.e11 = contribThis * self.e11 + contribOp * op.e11
		out.e12 = contribThis * self.e12 + contribOp * op.e12
		out.e13 = contribThis * self.e13 + contribOp * op.e13
		out.e21 = contribThis * self.e21 + contribOp * op.e21
		out.e22 = contribThis * self.e22 + contribOp * op.e22
		out.e23 = contribThis * self.e23 + contribOp * op.e23
		out.e31 = contribThis * self.e31 + contribOp * op.e31
		out.e32 = contribThis * self.e32 + contribOp * op.e32
		out.e33 = contribThis * self.e33 + contribOp * op.e33
	
		return out

e = MyEffect()
e.affect()
